/*
 * KeyBoard IR receiver v0.1 - (c) 2002 by Juggler@CyberFlat.nl
 *
 */

#include "defs.h"

#include <io.h>
#include <signal.h>
#include <interrupt.h>
#include <progmem.h>

typedef unsigned char  u08;
typedef          char  s08;
typedef unsigned short u16;
typedef          short s16;

#include "timer.h"
#include "fifo.c"
#include "keys.h"

#define pc_sbi(PORT, VALUE) outp((inp(PORT) |  VALUE), PORT)
#define pc_cbi(PORT, VALUE) outp((inp(PORT) & ~VALUE), PORT)

// Pre-defs
void kb_fetch(u08 data);
void kb_decode(u08 data);
void kb_init(u08 n);
void pc_decode(u08 mouse, u08 data);
void pc_transmit(u08 mouse, u08 data);

// Global PC/KB variables
static u08 pc_last_send = 0;	// Last byte send to pc
static u08 pc_last_receive = 0;	// Last byte received from pc
//static u08 pc_mouse = 0;	// Default keyboard

static u08 kb_state = 0;
static u08 kb_bitcnt = 0;
static u08 kb_byte = 0;
static u08 kb_counter = 0;
static u08 kb_cmd = 0;

// Signal handler for incomming pulse
SIGNAL(SIG_INTERRUPT0) {
  // re-sync timer
  outp(KB_TCNT_HALF, TCNT0);

  // check for mega-space (start-bit)
  if(kb_counter < 32) return;

  // fake 5 zero-bits already came in
  kb_state = 1;
  kb_bitcnt = 5;
  kb_byte = 0;
  kb_counter = 0;

  // turn off power led
  sbi(LED_PORT, LED_POWER);
}

void kb_lookup(u08 byte, u08 *key, u08 *extended) {
}

// Signal handler for timer interrupt
SIGNAL(SIG_OVERFLOW0) {
  u08 key, extended;

  // reset timer
  outp(KB_TCNT, TCNT0);

  // increase counter if possible
  if(kb_counter < 0xFF) kb_counter++;

  // do anything at all?
  if(! kb_state) return;

  // shift data in
  kb_byte <<= 1;
  if((inp(PIND) & (1<<2)) == 0) kb_byte |= 1;

  // check for complete byte
  if(++kb_bitcnt < 8) return;

  // take action
  switch(kb_state) {
    case(1):
      // check for valid start byte
      if(kb_byte == KB_START)
        kb_state++;
      else
        kb_state = 0;

      break;
    case(2):
      // check for valid command
      if(kb_byte == KB_UP || kb_byte == KB_DOWN)
        kb_state++;
      else
        kb_state = 0;

      // store command anyway
      kb_cmd = kb_byte;

      break;
    case(3):
      // reset state
      kb_state = 0;

      // turn power led back on
      cbi(LED_PORT, LED_POWER);

      // check for valid range
      if(kb_byte < 128) break;

      // look-up key
      //kb_lookup(kb_byte - 128, &key, &extended);
      kb_byte -= 128;
      key = __lpm_macro(&kb_keys[kb_byte].key);
      extended = __lpm_macro(&kb_keys[kb_byte].extended);

      // check for valid key
      if(key == 0) break;

      // store extended scancode
      if(extended) fifo_store(0xE0);

      // store release scancode
      if(kb_cmd == KB_UP) fifo_store(0xF0);

      // store key scancode
      fifo_store(key);

      break;
  }

  // reset vars 
  kb_bitcnt = 0;
  kb_byte = 0;
}

// Transmit a bit to a host PC
void pc_transmit_bit(u08 match_clock, u08 match_data, u08 bit) {
  if(bit != 0)	// '1'
    pc_cbi(PC_DDR_DATA, match_data);
  else		// '0'
    pc_sbi(PC_DDR_DATA, match_data);

  timer_delay(20);
  pc_sbi(PC_DDR_CLOCK, match_clock);	// Bring clock low
  timer_delay(40);
  pc_cbi(PC_DDR_CLOCK, match_clock);	// Release clock
  timer_delay(20);
}

// Transmit a byte to a host PC
void pc_transmit(u08 mouse, u08 data) {
  u08 match_data = 0;
  u08 match_clock = 0;
  u08 parity = 0;
  u08 i;

  // Fetch pin layout
  if(mouse) {
    match_data  = (1 << PC_MOUSE_DATA);
    match_clock = (1 << PC_MOUSE_CLOCK);
  } else {
    match_data  = (1 << PC_KEY_DATA);
    match_clock = (1 << PC_KEY_CLOCK);
  }

  // Wait for released clock
  while((inp(PC_PIN_CLOCK) & match_clock) == 0) { }

  // Check for low data line
  if((inp(PC_PIN_DATA) & match_data) == 0) return;

  // Transmit start bit
  pc_transmit_bit(match_clock, match_data, 0);

  // Transmit 8 data bits
  for(i = 0; i < 8; i++) {
    pc_transmit_bit(match_clock, match_data, data & 0x01);
    parity += (data & 0x01);
    data >>= 1;
  }

  // Transmit parity bit
  pc_transmit_bit(match_clock, match_data, 1 - (parity & 0x01));

  // Transmit stop bit
  pc_transmit_bit(match_clock, match_data, 1);

  // Store last byte send
  pc_last_send = data;

  timer_delay(30);
}

void leds_set(u08 value) {
  // Update SCROLL lock
  if(value & 1)
    cbi(LED_PORT, LED_SCROLL);
  else
    sbi(LED_PORT, LED_SCROLL);

  // Update NUM lock
  if(value & 2)
    cbi(LED_PORT, LED_NUM);
  else
    sbi(LED_PORT, LED_NUM);
  
  // Update Caps lock
  if(value & 4)
    cbi(LED_PORT, LED_CAPS);
  else
    sbi(LED_PORT, LED_CAPS);
}

// Decode received byte from host PC mouse
void pc_decode_mouse(u08 data) {
  // Walk thru main byte
  // ... bla bla ...
}

// Decode received byte from host PC keyboard
void pc_decode_key(u08 data) {
  u08 nak = 0;

  // Walk thru main byte
  switch(data) {
    case(0xFF):	// Reset
      pc_transmit(0, 0xAA);	// Self-test passed ;-)
      pc_transmit(0, 0xFA);	// Ack
      break;
    case(0xFE):	// Resend character
      pc_transmit(0, pc_last_send);
      break;
/*
    case(0xFA):	// Set all keys typematic/make/break
    case(0xF9):	// Set all keys make
    case(0xF8):	// Set all keys make/break
    case(0xF7):	// Set all keys typematic
    case(0xF6):	// Set default
      pc_transmit(0, 0xFA);     // Ack
      break;
*/
    case(0xF5):	// Disable
      pc_transmit(0, 0xFA);     // Ack
      break;
    case(0xF4):	// Enable
      pc_transmit(0, 0xFA);     // Ack
      break;
    case(0xF3):	// Set typematic rate/delay
      pc_transmit(0, 0xFA);     // Ack
      break;
    case(0xF2):	// Read ID
      pc_transmit(0, 0xFA);     // Ack
      pc_transmit(0, 0x83);     // ID Low
      pc_transmit(0, 0xAB);     // ID High
      break;
    case(0xED):	// Set/Reset LEDs
      leds_set(0);
      pc_transmit(0, 0xFA);     // Ack
      break;
    default:
      nak++;
  }

  // Walk thru control byte
  switch(pc_last_receive) {
    case(0xED):	// Set/Reset LEDs
      leds_set(data);
      pc_transmit(0, 0xFA);     // Ack
      break;
    case(0xF3):	// Set typematic rate/delay
      pc_transmit(0, 0xFA);     // Ack
      break;
    case(0xF0):	// Set scancode set
      pc_transmit(0, 0xFA);     // Ack
      break;
    default:
      nak++;
  }

  // Need to NAK?
  if(nak == 2) pc_transmit(0, 0xFE);

  // Store last byte received
  pc_last_receive = data;
}

// Reveive a bit from host PC
u08 pc_receive_bit(u08 match_clock, u08 match_data) {
  timer_delay(20);
  pc_sbi(PC_DDR_CLOCK, match_clock); 	// Bring clock low
  timer_delay(40);
  pc_cbi(PC_DDR_CLOCK, match_clock);	// Release clock
  timer_delay(20);

  // Read and return data
  if(inp(PC_PIN_DATA) & match_data)
    return(1);

  return(0);
}

// Listen to host PC, receive a byte from host PC
void pc_receive(u08 mouse) {
  u08 match_data = 0;
  u08 match_clock = 0;
  u08 i, parity;
  u08 data = 0;

  // Fetch pin layout
  if(mouse) {
    match_data  = (1 << PC_MOUSE_DATA);
    match_clock = (1 << PC_MOUSE_CLOCK);
  } else {
    match_data  = (1 << PC_KEY_DATA);
    match_clock = (1 << PC_KEY_CLOCK);
  }

  // Check if clock has been brought low by the host PC
  if((inp(PC_PIN_CLOCK) & match_clock) != 0) return;

  // Check if data has been brought low by the host PC
  if((inp(PC_PIN_DATA) & match_data) != 0) return;

  // Wait for host to release clock
  while((inp(PC_PIN_CLOCK) & match_clock) == 0) { }

  // Check if data still low -- No error occured, abort.
  if((inp(PC_PIN_DATA) & match_data) != 0) return;

  // Receive data bits
  for(i = 0; i < 8; i++) {
    data >>= 1;

    if(pc_receive_bit(match_clock, match_data)) data |= 0x80;
  }

  // Receive parity
  parity = pc_receive_bit(match_clock, match_data);

  // Receive stop bit
  pc_receive_bit(match_clock, match_data);

  // Generate acknowledge bit
  timer_delay(15);
  pc_sbi(PC_DDR_DATA, match_data);
  timer_delay(5);
  pc_sbi(PC_DDR_CLOCK, match_clock);
  timer_delay(40);
  pc_cbi(PC_DDR_CLOCK, match_clock);
  timer_delay(5);
  pc_cbi(PC_DDR_DATA, match_data);

  // Check parity bit. -- Generate an error if parity bit is incorrect

  // Give host some time?
  timer_delay(45);

  // Decode data byte
  if(! mouse)
    pc_decode_key(data);
  else
    pc_decode_mouse(data);
}

void delay(u16 ms) {
  u16 i;

  // Delay ms times
  for(i = 0; i < ms; i++) timer_delay(1000);
}

// Enter BAT
void bat(void) {
  // Turn off leds and delay 200 ms * 3 = 600 ms
  delay(200);
  sbi(LED_PORT,LED_NUM);
  delay(200);
  sbi(LED_PORT,LED_CAPS);
  delay(200);
  sbi(LED_PORT,LED_SCROLL);

  // Self-test passed ;-)
  pc_transmit(0, 0xAA);
}

// Global initialization routine
void init(void) {
  cli();					// Disable global interrupts

  // I/O
  outp(0x00, DDRB);				// Setup PortB for input
  outp(0x00, PORTB);
  outp(0x00, DDRD);				// Setup PortD for input
  outp(0x00, PORTD);

  // Setup LED outputs
  outp((1<<LED_POWER)|(1<<LED_NUM)|(1<<LED_CAPS)|(1<<LED_SCROLL), LED_DDR);
  outp(0x00, LED_PORT);

  // UART
  outp((u08)UART_BAUD_SLCT, UBRR);		// Set UART buadrate
  outp((1<<TXEN), UCR);				// Setup UART interrupt

  // Initialize timer
  timer_init();

  // Initialize timer interrupt
  outp((1<<CS01)|(1<<CS00), TCCR0);
  outp((1<<TOIE0), TIMSK);			// Enable timer0 clocked at CK/64
  outp(KB_TCNT, TCNT0);

  // Initialize external interrupt
  outp((1<<ISC01), MCUCR);			// Falling edge of INT0 generates an irq
  outp((1<<INT0), GIMSK);			// enable irq0

  sei();					// Enable global interrupts
}

int main(void) {
  u08 byte;

  // Global initialization
  init();

  // Do a BAT (self test)
  bat();

  // Live forever
  for(;;) {
    // Process byte from keyboard fifo
    if(! fifo_load(&byte)) pc_transmit(0, byte);

    // Listen to PC
    pc_receive(0);	// keyboard
  }
}
